from concurrent.futures import Future

from .common import ResponseFuture


def wrap_future(future: ResponseFuture) -> Future:
    """
    Convert a Cassandra's future to a normal Python future.
    The returned future will be marked as running.

    future is returned when it's already a Python future.

    :param future: cassandra future to wrap
    :return: a standard Python future
    """
    if isinstance(future, Future):
        return future

    fut = Future()
    fut.set_running_or_notify_cancel()
    future.add_callback(lambda result: fut.set_result(result))
    future.add_errback(lambda exception: fut.set_exception(exception))
    return fut



import typing as tp

TIME_MODIFIERS = [
    ('s', 1),
    ('m', 60),
    ('h', 60 * 60),
    ('d', 24 * 60 * 60),
    ('w', 7 * 24 * 60 * 60)
]


def parse_time_string(s: tp.Union[int, float, str]) -> float:
    """
    Parse a time string into seconds, so eg. '30m' will be equal to 1800, and so will
    be '30 min'.

    This will correctly parse:
    - seconds
    - minutes
    - hours
    - days
    - weeks

    .. warning:: This does not handle fractions of a second!

    :param s: time string or time value in seconds
    :return: value in seconds
    """
    if isinstance(s, (int, float)):
        return s

    for modifier, multiple in TIME_MODIFIERS:
        if modifier in s:
            return float(s[:s.index(modifier)]) * multiple

    return float(s)



import codecs
import os
import re
import typing as tp

from satella.exceptions import ConfigurationValidationError
from .base import Descriptor, ConfigDictValue
from .registry import register_custom_descriptor


@staticmethod
def _make_boolean(v: tp.Any) -> bool:
    if isinstance(v, str):
        if v.upper() == 'TRUE':
            return True
        elif v.upper() == 'FALSE':
            return False
        else:
            raise ConfigurationValidationError('Unknown value of "%s" posing to be a bool'
                                               % (v,))
    else:
        return bool(v)


@register_custom_descriptor('bool')
class Boolean(Descriptor):
    """
    This value must be a boolean, or be converted to one
    """
    BASIC_MAKER = _make_boolean


@register_custom_descriptor('int')
class Integer(Descriptor):
    """
    This value must be an integer, or be converted to one
    """
    BASIC_MAKER = int


@register_custom_descriptor('float')
class Float(Descriptor):
    """
    This value must be a float, or be converted to one
    """
    BASIC_MAKER = float


@register_custom_descriptor('str')
class String(Descriptor):
    """
    This value must be a string, or be converted to one
    """
    BASIC_MAKER = str


class FileObject:
    """
    What you get for values in schema of :class:`~satella.configuration.schema.File`.

    This object is comparable and hashable, and is equal to the string of it's path
    """
    __slots__ = 'path',

    def __init__(self, path: str):
        self.path = path

    def __repr__(self):
        return '<File object %s>' % (self.path,)

    def __str__(self):
        return self.path

    def __eq__(self, other) -> bool:
        return self.path == str(other) and isinstance(other, FileObject)

    def __hash__(self) -> int:
        return hash(self.path)

    def get_value(self, encoding: tp.Optional[str] = None) -> tp.Union[str, bytes]:
        """
        Read in the entire file into memory

        :param encoding: optional encoding to apply. If None given, bytes will be returned
        :return: file contents
        """
        with open(self.path, 'rb') as f_in:
            data = f_in.read()
        if encoding:
            return data.decode(encoding)
        else:
            return data

    def open(self, mode: str):
        """
        Open the file in specified mode

        :param mode: mode to open the file in
        :return: file handle
        """
        return open(self.path, mode)


class DirectoryObject:
    """
    What you get for values in schema of :class:`~satella.configuration.schema.Directory`.

    This object is comparable and hashable, and is equal to the string of it's path
    """
    __slots__ = 'path',

    def __init__(self, path: str):
        self.path = path

    def __repr__(self):
        return '<Directory object %s>' % (self.path,)

    def __str__(self):
        return self.path

    def __eq__(self, other) -> bool:
        return self.path == str(other) and isinstance(other, DirectoryObject)

    def __hash__(self) -> int:
        return hash(self.path)

    def get_files(self) -> tp.Iterable[str]:
        """
        Return a list of files inside this directory
        :return:
        """
        return os.listdir(self.path)


@staticmethod
def _make_file(v: str) -> bool:
    if not os.path.isfile(v):
        raise ConfigurationValidationError('Expected to find a file under %s'
                                           % (v,))
    return FileObject(v)


@register_custom_descriptor('file')
class File(Descriptor):
    """
    This value must be a valid path to a file. The value in your schema will be
    an instance of :class:`~satella.configuration.schema.basic.FileObject`
    """

    BASIC_MAKER = _make_file


@register_custom_descriptor('file_contents')
class FileContents(Descriptor):
    """
    This value must be a valid path to a file. The value in your schema will be
    the contents of this file, applied with encoding (if given). By default, bytes will be read in
    """

    def __init__(self, encoding: tp.Optional[str] = None, strip_afterwards: bool = False):
        super().__init__()
        self.encoding = encoding
        self.strip_afterwards = strip_afterwards

    def BASIC_MAKER(self, c: str):
        if not self.encoding:
            with open(c, 'rb') as f_in:
                y = f_in.read()
        else:
            with codecs.open(c, 'r', encoding=self.encoding) as f_in:
                y = f_in.read()

        if self.strip_afterwards:
            y = y.strip()
        return y


@staticmethod
def _make_directory(v: str) -> bool:
    if not os.path.isdir(v):
        raise ConfigurationValidationError('Expected to find a directory under %s'
                                           % (v,))
    return DirectoryObject(v)


@register_custom_descriptor('dir')
class Directory(Descriptor):
    """
    This value must be a valid path to a file. The value in your schema will be
    an instance of :class:`~satella.configuration.schema.basic.FileObject`
    """

    BASIC_MAKER = _make_directory


class Regexp(String):
    """
    Base class for declaring regexp-based descriptors. Overload it's attribute REGEXP. Use as
    following:

    >>> class IPv6(Regexp):
    >>>     REGEXP = '(([0-9a-f]{1,4}:)' ...
    """
    __slots__ = ('regexp',)

    REGEXP = r'.*'

    def __init__(self):
        super().__init__()
        if isinstance(self.REGEXP, str):
            self.regexp = re.compile(self.REGEXP)
        else:
            self.regexp = self.REGEXP

    def __call__(self, value: ConfigDictValue) -> str:
        value = super(Regexp, self).__call__(value)

        match = self.regexp.match(value)
        if not match:
            raise ConfigurationValidationError('value does not match %s' % (self.REGEXP.pattern,),
                                               value)

        return match.group(0)

    def __str__(self):
        return 'Regexp(%s)' % (self.REGEXP.pattern,)


@register_custom_descriptor('ipv4')
class IPv4(Regexp):
    """
    This must be a valid IPv4 address (no hostnames allowed)
    """
    REGEXP = r'\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3}'



import logging
import enum
import gc
import unittest

from satella.coding import update_key_if_not_none, overload, class_or_instancemethod, \
    update_key_if_true, get_arguments, call_with_arguments, chain_callables, Closeable, \
    contains, enum_value, for_argument
from satella.coding.structures import HashableMixin, ComparableEnum
from satella.coding.transforms import jsonify, intify

logger = logging.getLogger(__name__)

class TestCase(unittest.TestCase):

    def test_closeable_semi_initialized_classes(self):
        slf = self

        class Test(Closeable):
            def __init__(self):
                raise ValueError()
                self.a = 5
                super().__init__()

            def close(self):
                if super().close():
                    slf.fail('Tried to close an uninitialized class!')

        self.assertRaises(ValueError, Test)

    def test_cant_compare_me(self):
        class Uncomparable:
            def __eq__(self, other):
                raise TypeError()

            def __str__(self):
                return 'test'

        @for_argument(y=str)
        def test(y = Uncomparable()):
            return y

        b = test()
        self.assertEqual(b, 'test')

    def test_contains(self):
        class CEnum(ComparableEnum):
            A = 1
            B = 2

        a = [CEnum.A, CEnum.B]
        self.assertTrue(contains(2, a))
        self.assertFalse(contains(3, a))

        self.assertEqual(enum_value(CEnum.A), 1)
        self.assertEqual(enum_value(1), 1)

    def test_closeable(self):
        a = {'test': False}

        class MyClose(Closeable):
            def close(self):
                if super().close():
                    a['test'] = True

        b = MyClose()
        self.assertFalse(b.closed)
        del b
        gc.collect()
        self.assertTrue(a['test'])

    def test_chain_callables(self):
        def a():
            return 5

        def mul2(b):
            return b*2

        self.assertEqual(chain_callables(a, mul2)(), 10)

        def two():
            return 2

        self.assertEqual(chain_callables(a, two)(), 2)

    def test_hashable_mixin(self):
        class MixedIn(HashableMixin):
            pass

        a = MixedIn()
        b = MixedIn()

        self.assertNotEqual(hash(a), hash(b))
        self.assertNotEqual(a, b)

    def test_intify(self):
        self.assertEqual(intify(None), 0)
        self.assertEqual(intify('2'), 2)
        self.assertRaises(ValueError, lambda: intify(object()))
        self.assertEqual(intify([1, 2, 3]), 3)

    def test_execute_with_locals(self):
        def fun(a, b, *args, c=None, **kwargs):
            if len(kwargs):
                d = kwargs[next(iter(kwargs))]
            else:
                d = 0
            return a + b + args[0] + c + d

        self.assertEqual(call_with_arguments(fun, {
            'a': 5,
            'b': 5,
            'args': (5,),
            'c': 5
        }), 20)
        self.assertEqual(call_with_arguments(fun, {
            'a': 5,
            'b': 5,
            'args': (5,),
            'kwargs': {
                'd': 10,
            },
            'c': 5
        }), 30)
        self.assertRaises(TypeError, lambda: call_with_arguments(fun, {}))
        self.assertRaises(ValueError, lambda: call_with_arguments(fun, {
            'a': 5,
            'b': 5,
            'args': (5,),
            'd': 10,
            'c': 5
        }))

    def test_get_arguments(self):
        def fun(a, b, *args, c=None, **kwargs):
            ...

        self.assertRaises(TypeError, lambda: get_arguments(fun))
        self.assertEqual(get_arguments(fun, 3, 4, 5, 6, c=4, d=5),
                         {
                             'a': 3,
                             'b': 4,
                             'args': (5, 6),
                             'c': 4,
                             'kwargs': {
                                 'd': 5
                             }
                         })

        def fun(a=5, c=7, *args, d=8, **kwargs):
            ...

        self.assertEqual(get_arguments(fun), {
            'a': 5,
            'c': 7,
            'args': (),
            'd': 8,
            'kwargs': {}
        })

        def fun(a=6):
            ...

        self.assertEqual(get_arguments(fun, 7),
                         {'a': 7})

    def test_jsonify(self):
        def iterate():
            yield 5

        class StrFails:
            def __str__(self):
                raise TypeError()

        class Enu(enum.IntEnum):
            A = 0

        self.assertEqual(jsonify(Enu.A), 0)
        self.assertEqual(jsonify(iterate()), [5])
        self.assertEqual(jsonify([iterate(), iterate()]), [[5], [5]])
        self.assertEqual(jsonify({'5': iterate()}), {'5': [5, ]})
        self.assertTrue(isinstance(jsonify(StrFails()), str))

    def test_update_key_if_true(self):
        a = {5: False}
        update_key_if_true(a, 5, True, False)
        self.assertFalse(a[5])
        update_key_if_true(a, 5, True, True)
        self.assertTrue(a[5])
        update_key_if_true(a, 6, True)
        self.assertTrue(a[6])

    def test_class_or_instancemethod(self):

        a = {}

        class MyClass:
            @class_or_instancemethod
            def method(self_or_cls):
                if isinstance(self_or_cls, MyClass):
                    a['method'] = True
                else:
                    a['classmethod'] = True

        b = MyClass()
        b.method()
        self.assertEqual(a, {'method': True})
        MyClass.method()
        self.assertEqual(a, {'method': True, 'classmethod': True})

    def test_overload(self):
        a = {}

        @overload
        def what_type(x: str):
            a['type'] = 'str'

        @what_type.overload
        def what_type(x: int):
            a['type'] = 'int'

        what_type('test')
        self.assertEqual(a['type'], 'str')
        what_type(2)
        self.assertEqual(a['type'], 'int')
        self.assertRaises(TypeError, lambda: what_type(2.0))

    def test_update_key_if_not_none(self):
        a = {}
        update_key_if_not_none(a, 'test', None)
        update_key_if_not_none(a, 'test2', 2)
        self.assertEqual(a, {'test2': 2})
        update_key_if_not_none(a, {'test3': 3, 'test4': None})
        self.assertEqual(a, {'test2': 2, 'test3': 3})



import stat
import typing as tp
import os
import sys
import warnings

from satella.coding import silence_excs


def whereis(name: str) -> tp.Iterator[str]:
    """
    Looking in PATH return a sequence of executable files having provided name.

    Additionally, on Windows, it will use PATHEXT.

    .. note:: on Windows name is supposed to be without extension!

    :param name: name of the executable to search for
    :return: an iterator of absolute paths to given executable
    """
    if sys.platform.startswith('win'):
        paths_to_look_in = os.environ.get('PATH', '').split(';')
        name = name.upper()
        available_extensions = os.environ.get('PATHEXT', '.com;.bat;.exe').upper().split(';')
    else:
        paths_to_look_in = os.environ.get('PATH', '').split(':')
        available_extensions = '',

    for directory in paths_to_look_in:
        with silence_excs(FileNotFoundError):
            for file in os.listdir(directory):
                path = os.path.join(directory, file)
                if 'x' not in stat.filemode(os.stat(path).st_mode):
                    continue

                if sys.platform.startswith('win'):  # a POSIX-specific check
                    file = file.upper()  # paths are not case-sensitive on Windows

                for extension in available_extensions:
                    if file == '%s%s' % (name, extension):
                        yield path


def is_running_as_root() -> bool:
    """
    Is this process running as root?

    Checks whether effective UID is 0

    :return: bool
    :raises OSError: called on Windows!
    """

    if sys.platform.startswith('win'):
        raise OSError('Routine unavailable on Windows!')

    return os.geteuid() == 0


def suicide(kill_entire_pg: bool = True) -> None:
    """
    Kill self.

    :param kill_entire_pg: whether to kill entire PG if a session leader. Won't work on Windows.
    """
    my_pid = os.getpid()
    if sys.platform.startswith('win'):
        if kill_entire_pg:
            warnings.warn('Windows does not support process leader groups', RuntimeWarning)
        os.kill(my_pid, -9)

    if kill_entire_pg and os.getpgid(0) == my_pid:
        os.killpg(my_pid, 9)
    else:
        os.kill(my_pid, 9)



from .arguments import auto_adapt_to_methods, attach_arguments, for_argument, \
    execute_before, copy_arguments, replace_argument_if, transform_result, \
    transform_arguments, execute_if_attribute_none, execute_if_attribute_not_none, \
    cached_property
from .decorators import wraps, chain_functions, has_keys, short_none, memoize, return_as_list, \
    default_return, cache_memoize, call_method_on_exception
from .flow_control import loop_while, queue_get
from .preconditions import postcondition, precondition
from .retry_dec import retry

__all__ = ['retry', 'transform_result', 'transform_arguments',
           'execute_before', 'postcondition', 'precondition', 'wraps', 'queue_get',
           'chain_functions', 'has_keys', 'short_none', 'auto_adapt_to_methods',
           'attach_arguments', 'for_argument', 'loop_while', 'memoize',
           'copy_arguments', 'replace_argument_if', 'return_as_list',
           'default_return', 'cache_memoize', 'call_method_on_exception',
           'execute_if_attribute_none', 'execute_if_attribute_not_none',
           'cached_property']



import time
import typing as tp

__all__ = ['time_as_int', 'time_ms', 'sleep', 'time_us', 'ExponentialBackoff']

from satella.coding.concurrent.thread import Condition
from satella.exceptions import WouldWaitMore
from .parse import parse_time_string
from .measure import measure


def sleep(y: tp.Union[str, float], abort_on_interrupt: bool = False) -> bool:
    """
    Sleep for given interval.

    This won't be interrupted by KeyboardInterrupted, and always will sleep for given time interval.
    This will return at once if x is negative

    :param y: the interval to wait in seconds, can be also a time string
    :param abort_on_interrupt: whether to abort at once when KeyboardInterrupt is seen
    :returns: whether the function has completed its sleep naturally. False is seen on
        aborts thanks to KeyboardInterrupt only if abort_on_interrupt is True
    """
    y = parse_time_string(y)
    if y < 0:
        return

    with measure() as measurement:
        while measurement() < y:
            try:
                time.sleep(y - measurement())
            except KeyboardInterrupt:
                if abort_on_interrupt:
                    return False
    return True


def time_as_int() -> int:
    """
    Syntactic sugar for

    >>> from time import time
    >>> int(time())
    """
    return int(time.time())


def time_ms() -> int:
    """
    Syntactic sugar for

    >>> from time import time
    >>> int(time()*1000)

    This will try to use time.time_ns() if available
    """
    try:
        return time.time_ns() // 1000000
    except AttributeError:
        return int(time.time() * 1000)


def time_us() -> int:
    """
    Syntactic sugar for

    >>> from time import time
    >>> int(time()*1000000)

    This will try to use time.time_ns() if available
    """
    try:
        return time.time_ns() // 1000
    except AttributeError:
        return int(time.time() * 1000000)





import io
import os
import pickle
import tempfile
import logging
import threading

from satella.coding import silence_excs
from satella.exception_handling import DumpToFileHandler, exception_handler
from satella.instrumentation import Traceback
from . import ExceptionHandlingTestCase


logger = logging.getLogger(__name__)


class TestDumpToFile(ExceptionHandlingTestCase):

    def test_exception_handler(self):

        test = False

        def raise_exception():
            raise KeyError()

        @exception_handler(priority=0)
        def handle_exception(type_, value, traceback):
            nonlocal test
            test = True
            return True

        handle_exception.install()

        try:
            t = threading.Thread(target=raise_exception)
            t.start()
            t.join()

            self.assertTrue(test)
        finally:
            handle_exception.uninstall()

    def setUp(self):
        self.sq, self.sa, self.tf = io.StringIO(), io.StringIO(), tempfile.mktemp()
        self.op = io.BytesIO()
        self.exception_handler = DumpToFileHandler(
            human_readables=[self.sq, self.sa, self.tf, None, logger, (logger, logging.WARNING)],
            trace_pickles=[self.op])
        self.exception_handler.install()

    def tearDown(self):
        self.exception_handler.uninstall()

    def test_dump_to_file(self):
        try:
            def makeexc():
                print(hello)  # raises NameError

            self.make_exception(makeexc)

            self.sq, self.sa = self.sq.getvalue(), self.sa.getvalue()
            with open(self.tf, 'r', encoding='utf8') as fi:
                sb = fi.read()

            def assertIn(txt):
                self.assertIn(txt, self.sq)
                self.assertIn(txt, self.sa)
                self.assertIn(txt, sb)

            assertIn('NameError')
            assertIn('hello')
        finally:
            with silence_excs(FileNotFoundError):
                os.unlink(self.tf)

        tbp = pickle.loads(self.op.getvalue())
        self.assertIsInstance(tbp, Traceback)



class EmptyContextManager:
    """
    A context manager that does nothing. Only to support conditional change of context managers,
    eg in such a way:

    >>> if tx_enabled:
    >>>     ctxt = transaction.atomic
    >>> else:
    >>>     ctxt = EmptyContextManager()
    >>> with ctxt:
    >>>     ...
    """
    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        return False





import logging
import queue
import unittest
from socket import socket

import time
from satella.coding import wraps, chain_functions, postcondition, \
    log_exceptions, queue_get, precondition, short_none
from satella.coding.decorators import auto_adapt_to_methods, attach_arguments, \
    execute_before, loop_while, memoize, copy_arguments, replace_argument_if, \
    retry, return_as_list, default_return, transform_result, transform_arguments, \
    cache_memoize, call_method_on_exception, execute_if_attribute_none, \
    execute_if_attribute_not_none, cached_property
from satella.coding.predicates import x
from satella.exceptions import PreconditionError

logger = logging.getLogger(__name__)


class TestDecorators(unittest.TestCase):

    def test_cached_property(self):
        class Example:
            def __init__(self):
                self._a = None
                self.called = 0

            @property
            @cached_property('_a')
            def a(self):
                self.called += 1
                return 'abc'
        a = Example()
        self.assertEqual(a.a, 'abc')
        self.assertEqual(a.called, 1)
        self.assertEqual(a.a, 'abc')
        self.assertEqual(a.called, 1)

    def test_execute_if_attribute_not_none(self):
        class ExecIfAttrNone:
            def __init__(self):
                self.attr = None
                self.called = 0

            @execute_if_attribute_not_none('attr')
            def run_me(self):
                self.called += 1

        e = ExecIfAttrNone()
        self.assertEqual(e.called, 0)
        e.run_me()
        self.assertEqual(e.called, 0)
        e.attr = 2
        e.run_me()
        self.assertEqual(e.called, 1)

    def test_execute_if_attribute_none(self):
        class ExecIfAttrNone:
            def __init__(self):
                self.attr = None
                self.called = 0

            @execute_if_attribute_none('attr')
            def run_me(self):
                self.called += 1

        e = ExecIfAttrNone()
        self.assertEqual(e.called, 0)
        e.run_me()
        self.assertEqual(e.called, 1)
        e.attr = 2
        e.run_me()
        self.assertEqual(e.called, 1)

    def test_call_method_on_exception(self):
        a = {'called': False}
        slf = self

        class Test:

            def close(self, b):
                nonlocal a
                slf.assertEqual(b, 2)
                a['called'] = True

            @call_method_on_exception(ValueError, 'close', 2)
            def do(self):
                raise ValueError()

        c = Test()
        c.do()
        self.assertTrue(a['called'])

    def test_cached_memoizer(self):
        a = {'calls': 0}

        @cache_memoize(1)
        def returns(b):
            nonlocal a
            a['calls'] += 1
            return b

        self.assertEqual(returns(6), 6)
        self.assertEqual(a['calls'], 1)
        self.assertEqual(returns(6), 6)
        self.assertEqual(a['calls'], 1)
        time.sleep(1.1)
        self.assertEqual(returns(6), 6)
        self.assertEqual(a['calls'], 2)

    def test_transform_arguments(self):
        @transform_arguments(a='a*a')
        def square(a):
            return a

        self.assertEqual(square(4), 16)

    def test_transform_result(self):
        @transform_result('x*a')
        def square(a):
            return a

        self.assertEqual(square(4), 16)

    def test_default_returns(self):
        @default_return(6)
        def returns(v):
            return v

        self.assertEqual(returns(None), 6)
        self.assertEqual(returns(4), 4)

    def test_return_as_list(self):
        @return_as_list(ignore_nulls=True)
        def test():
            yield 2
            yield 3
            yield None
            yield 4

        self.assertEqual(test(), [2, 3, 4])

        @return_as_list()
        def test():
            yield 2
            yield 3
            yield None
            yield 4

        self.assertEqual(test(), [2, 3, None, 4])

    def test_retry(self):
        a = {'test': 3, 'limit': 2, 'true': False, 'false': False,
             'retried': False}

        def on_failure(e):
            nonlocal a
            a['true'] = True

        def on_success(retries):
            nonlocal a
            a['false'] = True

        def on_retry(e):
            nonlocal a
            a['retried'] = True

        @retry(3, ValueError, swallow_exception=False, call_on_failure=on_failure,
               call_on_success=on_success,
               call_between_retries=on_retry)
        def do_op():
            a['test'] += 1
            if a['test'] < a['limit']:
                raise ValueError()

        do_op()
        self.assertTrue(a['false'])
        self.assertFalse(a['retried'])
        a = {'test': 0, 'limit': 2, 'true': False, 'false': False,
             'retried': False}

        do_op()
        self.assertTrue(a['retried'])
        self.assertTrue(a['false'])
        a['limit'] = 10
        a['false'] = False
        a['retried'] = False
        self.assertRaises(ValueError, do_op)
        self.assertTrue(a['true'])
        self.assertFalse(a['false'])
        self.assertTrue(a['retried'])
        self.assertTrue(a['retried'])

    def test_replace_argument_if(self):
        @replace_argument_if('y', x.int(), str)
        def ints_only(y):
            self.assertEqual(y, 2)
        ints_only(2)
        ints_only('2')

        @replace_argument_if('y', 1, predicate=x % 2 == 0)
        def only_odds(y):
            self.assertEqual(y % 2, 1)

        only_odds(0)
        only_odds(1)

        @replace_argument_if('args', (x[0], ), predicate=x.len() > 1)
        def args_len_1(*args):
            self.assertEqual(len(args), 1)

        args_len_1(1)
        args_len_1(1, 1)

    def test_copy_arguments(self):
        @copy_arguments()
        def alter_dict(dct):
            return dct.pop('a')

        b = {'a': 5}
        self.assertEqual(alter_dict(b), 5)
        self.assertEqual(b, {'a': 5})

    def test_memoize(self):
        a = {'call_count': 0}

        @memoize
        def memoizer(b):
            a['call_count'] += 1
            return b

        five = memoizer(5)
        self.assertEqual(a['call_count'], 1)
        five = memoizer(5)
        self.assertEqual(a['call_count'], 1)
        five = memoizer(6)
        self.assertEqual(a['call_count'], 2)

    def test_loop_while(self):
        class MyLooped:
            terminating = False
            i = 0

            @loop_while(x.i < 10)
            def run(self):
                self.i += 1

        a = MyLooped()
        a.run()
        self.assertGreaterEqual(a.i, 10)
        b = {'i': 0}

        @loop_while(lambda: b['i'] < 10)
        def run():
            nonlocal b
            b['i'] += 1

        run()
        self.assertGreaterEqual(b['i'], 10)

    def test_execute_before(self):
        a = 0

        @execute_before
        def increase_a(factor=1):
            nonlocal a
            a += factor

        @increase_a(factor=2)
        def launch_me():
            nonlocal a
            a += 1

        launch_me()
        self.assertEqual(a, 3)

    def test_precondition_none(self):
        @precondition(short_none('x == 2'))
        def x(y):
            return y
        x(2)
        x(None)
        self.assertRaises(PreconditionError, lambda: x(3))

    def test_queue_get(self):
        class Queue:
            def __init__(self):
                self.queue = queue.Queue()
                self.on_empty_called = False

            @queue_get('queue', timeout=0, method_to_execute_on_empty='process_on_empty')
            def process(self, item):
                pass

            def process_on_empty(self):
                self.on_empty_called = True

        q = Queue()
        q.queue.put(True)
        q.process()
        q.process()
        self.assertTrue(q.on_empty_called)

    def test_log_exceptions(self):
        try:
            with log_exceptions(logger):
                int('a')
        except ValueError:
            pass
        else:
            self.fail('exception swallowed!')

    def test_postcondition(self):
        @postcondition(lambda x: x == 2)
        def return_a_value(x):
            return x

        self.assertEqual(return_a_value(2), 2)
        self.assertRaises(PreconditionError, lambda: return_a_value(3))

    def test_auto_adapt_to_methods(self):
        @auto_adapt_to_methods
        def times_two(fun):
            def outer(a):
                return fun(a * 2)

            return outer

        class Test:
            @times_two
            def twice(self, a):
                return a * 2

        @times_two
        def twice(a):
            return a * 2

        self.assertEqual(Test().twice(2), 8)
        self.assertEqual(twice(2), 8)

    def test_chain_kwargs(self):
        @chain_functions
        def double_arguments(**kwargs):
            kwargs['a'] = kwargs['a'] * 2
            return kwargs

        @double_arguments
        def multiply_times_two(**kwargs):
            return kwargs['a'] * 2

        self.assertEqual(multiply_times_two(a=2), 8)

    def test_chain(self):
        @chain_functions
        def double_arguments(a):
            return a * 2

        @double_arguments
        def multiply_times_two(a):
            return a * 2

        self.assertEqual(multiply_times_two(2), 8)

    def test_attach_arguments(self):
        @attach_arguments(label=2)
        def test_me(**kwargs):
            self.assertEqual(kwargs, {'label': 2, 'value': 4})

        test_me(value=4)

    def test_wraps(self):
        @wraps(socket)
        class MySocket(socket):
            pass

        self.assertEqual(MySocket.__name__, socket.__name__)
        self.assertEqual(MySocket.__doc__, socket.__doc__)
        self.assertEqual(MySocket.__module__, socket.__module__)

    def test_wraps_onfunction(self):
        def my_fun(a):
            """Returns the argument"""
            return a

        @wraps(my_fun)
        def f(a):
            return my_fun(a)

        self.assertEqual(f.__doc__, my_fun.__doc__)
        self.assertEqual(f.__name__, my_fun.__name__)
        self.assertEqual(f.__module__, my_fun.__module__)



import os
import typing as tp
import unittest

from satella.coding import wraps
from satella.configuration.sources import BaseSource
from satella.exceptions import ConfigurationError


def mock_env(env: tp.Optional[str] = None, val: tp.Optional[str] = None) -> \
        tp.Callable[[tp.Callable], tp.Callable]:
    """
    Set an env and then clear it out
    :param env:
    :param val:
    :return:
    """

    def outer(fun: tp.Callable):
        @wraps(fun)
        def inner(*args, **kwargs):
            try:
                if env is not None:
                    os.environ[env] = val
                return fun(*args, **kwargs)
            finally:
                if env is not None:
                    del os.environ[env]

        return inner

    return outer


class SourceTestCase(unittest.TestCase):
    def assertSourceFails(self, source: BaseSource,
                          fails_with=ConfigurationError):
        self.assertRaises(fails_with, source.provide)

    def assertSourceEmpty(self, source: BaseSource):
        self.assertEqual(source.provide(), {})

    def assertSourceHas(self, source: BaseSource, value: tp.Any, postop=lambda x: x):
        self.assertEqual(postop(source.provide()), value)



import time
import unittest

from satella.instrumentation.cpu_time import calculate_occupancy_factor, sleep_cpu_aware, \
    CPUTimeAwareIntervalTerminableThread


class TestCPUTime(unittest.TestCase):
    def test_cpu_time_aware_terminable_thread_terminates(self):
        class TestingThread(CPUTimeAwareIntervalTerminableThread):
            def __init__(self):
                super().__init__('5s', 3, 0.5, terminate_on=ValueError)
                self.a = 0

            def loop(self) -> None:
                raise ValueError()

        TestingThread().start().terminate().join()

    def test_cpu_time_aware_terminable_thread(self):
        class TestingThread(CPUTimeAwareIntervalTerminableThread):
            def __init__(self):
                super().__init__('5s', 3, 0.5)
                self.a = 0

            def loop(self) -> None:
                self.a += 1

        tt = TestingThread()
        tt.start()
        time.sleep(0.2)
        self.assertEqual(tt.a, 1)
        time.sleep(5)
        self.assertEqual(tt.a, 2)
        tt.terminate().join()

    def test_sleep_except(self):
        c = time.monotonic()
        sleep_cpu_aware(1)
        self.assertGreaterEqual(time.monotonic() - c, 1)

    def test_calculate_occupancy_factor(self):
        c = calculate_occupancy_factor()
        self.assertGreaterEqual(c, 0)
        c = calculate_occupancy_factor()
        self.assertGreaterEqual(c, 0)
        for i in range(4):  # make at most 4 attempts
            if c < 1:
                break
            c = calculate_occupancy_factor()



import enum
import operator
import unittest
import base64

from satella.coding.predicates import x
from satella.coding.transforms import stringify, split_shuffle_and_join, one_tuple, \
    merge_series, pad_to_multiple_of_length, clip, b64encode, linear_interpolate, \
    hashables_to_int, none_if_false, merge_list, is_subset, unpack_dict


class TestTransforms(unittest.TestCase):

    def test_unpack_dict(self):
        a, b, c = unpack_dict({1: 2, 2: 3, 4: 5}, 1, 2, 4)
        self.assertTrue(a == 2 and b == 3 and c == 5)

        a, b, c = unpack_dict({1: 2, 2: 3, 4: 5}, 1, 2, 4,
                              map_through=x*2)
        self.assertTrue(a == 4 and b == 6 and c == 10)

    def test_is_subset(self):
        self.assertTrue(is_subset({}, {}))
        self.assertTrue(is_subset({}, {1: 2}))
        self.assertFalse(is_subset({2: 3}, {1: 2}))
        self.assertFalse(is_subset({1: 3}, {1: 2}))
        self.assertTrue(is_subset({1: 2}, {1: 2}))

    def test_merge_list(self):
        a = [(1, 1), (2, 2), (3, 3)]
        b = [(1, 2), (3, 2), (4, 4)]

        c = merge_list(a, b, merge_function=operator.add)
        self.assertEqual(list(c), [(1, 3), (2, 2),  (3, 5), (4, 4)])

        a = [(1, 1), (1, 3), (2, 2), (3, 3)]
        b = [(1, 2), (3, 2), (4, 4)]

        c = merge_list(a, b, merge_function=operator.add)
        self.assertEqual(list(c), [(1, 6), (2, 2),  (3, 5), (4, 4)])

        a = []
        b = []
        c = merge_list(a, b, merge_function=operator.add)
        self.assertEqual(list(c), [])

    def test_none_if_false(self):
        self.assertEqual(none_if_false(1), 1)
        self.assertEqual(none_if_false(''), None)

    def test_words_to_int(self):
        words = ['a', 'word', 'a']
        dct = hashables_to_int(words)
        self.assertEqual(len(dct), 2)
        self.assertEqual(len(list(dct.values())), 2)

    def test_linear_interpolate(self):
        self.assertEqual( linear_interpolate([(0, 10), (10, 20)], 5), 15)
        self.assertEqual( linear_interpolate([(0, 10), (10, 20)], 10), 20)
        self.assertEqual( linear_interpolate([(0, 10), (10, 20)], 2.5), 12.5)
        self.assertEqual( linear_interpolate([(0, 10), (10, 20)], 0), 10)
        self.assertEqual( linear_interpolate([(0, 10), (10, 20)], 10), 20)
        self.assertEqual( linear_interpolate([(0, 10), (10, 20), (20, 10)], 15), 15)
        self.assertRaises(ValueError, lambda: linear_interpolate([(0, 10), (10, 20)], 30))
        self.assertRaises(ValueError, lambda: linear_interpolate([(0, 10), (10, 20)], -2))
        self.assertEqual(linear_interpolate([(0, 10), (10, 20)], 30, True), 20)
        self.assertEqual(linear_interpolate([(0, 10), (10, 20)], -2, True), 10)

    def test_base64(self):
        a = b64encode(b'test')
        self.assertEqual(base64.b64decode(a), b'test')

    def test_empty_merge_series(self):
        sm = merge_series([], [])
        a = list(sm)
        self.assertFalse(a)

    def test_stringify_bug(self):
        self.assertEqual(stringify('test'), 'test')

    def test_clip(self):
        self.assertEqual(clip(5, 10, 15), 10)
        self.assertEqual(clip(25, 10, 15), 15)
        self.assertEqual(clip(12, 10, 15), 12)

    def test_pad_to_multiple_length(self):
        a = [1, 2, 3]
        a = pad_to_multiple_of_length(a, 16, None)
        self.assertTrue(len(a) % 16 == 0)

    def test_series_merger(self):
        s1 = [(100, 1), (200, 2), (300, 3)]
        s2 = ((100, 2), (200, 3), (300, 4))

        sm = merge_series(s1, s2)
        self.assertEqual(next(sm), (100, 1, 2))
        self.assertEqual(next(sm), (200, 2, 3))
        self.assertEqual(next(sm), (300, 3, 4))
        self.assertRaises(StopIteration, lambda: next(sm))

    def test_series_merger2(self):
        s1 = [(100, 1), (200, 2), (300, 3)]
        s2 = ((101, 2), (201, 3), (301, 4))

        sm = merge_series(s1, s2)
        self.assertEqual(next(sm), (101, 1, 2))
        self.assertEqual(next(sm), (200, 2, 2))
        self.assertEqual(next(sm), (201, 2, 3))
        self.assertEqual(next(sm), (300, 3, 3))
        self.assertEqual(next(sm), (301, 3, 4))
        self.assertRaises(StopIteration, lambda: next(sm))

    def test_series_merger_3(self):
        s1 = [(100, 1), (200, 2)]
        s2 = []

        sm = merge_series(s1, s2)

        self.assertRaises(StopIteration, lambda: next(sm))

    def test_series_merger_4(self):
        s1 = [(100, 1), (200, 2), (300, 3)]
        s2 = [(100, 1)]

        sm = merge_series(s1, s2)
        self.assertEqual(next(sm), (100, 1, 1))
        self.assertEqual(next(sm), (200, 2, 1))
        self.assertEqual(next(sm), (300, 3, 1))

        self.assertRaises(StopIteration, lambda: next(sm))

    def test_series_merger_5(self):
        s1 = [(100, 1), (200, 2), (300, 3)]
        sm = merge_series(s1)

        self.assertEqual(next(sm), (100, 1))
        self.assertEqual(next(sm), (200, 2))
        self.assertEqual(next(sm), (300, 3))
        self.assertRaises(StopIteration, lambda: next(sm))

    def test_merge_series(self):
        s1 = [(10, 'A'), (20, 'B')]
        s2 = [(15, 'C'), (17, 'D'), (21, 'E')]
        s3 = list(merge_series(s1, s2))
        self.assertEqual(s3, [(15, 'A', 'C'), (17, 'A', 'D'), (20, 'B', 'D'), (21, 'B', 'E')])

    def test_one_tuple(self):
        self.assertEqual(list(one_tuple([1, 2, 3])), [(1, ), (2, ), (3, )])

    def test_split_shuffle_and_join(self):
        x = [1, 2, 3, 4]
        x = split_shuffle_and_join(x, lambda y: y % 2 == 0)
        self.assertTrue(x == [1, 3, 2, 4] or x == [1, 3, 4, 2])

        x = [1, 2, 3, 4]
        x = split_shuffle_and_join(x, lambda y: y % 2 == 0, False)
        self.assertTrue(x == [2, 4, 1, 3] or x == [4, 2, 1, 3])

    def test_stringify(self):
        class Enum(enum.Enum):
            A = 0

        self.assertEqual(stringify(Enum.A), 'A')

        dct = {1: 2, 3: 4, 5: 6}
        self.assertEqual(stringify(dct), {'1': '2', '3': '4', '5': '6'})

        lst = [1, 2, 3, 4, 5]
        self.assertEqual(stringify(lst), ['1', '2', '3', '4', '5'])

        lst2 = [[1, 2, 3], 3, 4, 5]
        self.assertEqual(stringify(lst2), ['[1, 2, 3]', '3', '4', '5'])
        self.assertEqual(stringify(lst2, recursively=True), [['1', '2', '3'], '3', '4', '5'])

        dct2 = {1: [1, 2, 3]}
        self.assertEqual(stringify(dct2, recursively=True), {'1': ['1', '2', '3']})

        self.assertIsNone(stringify(None))
        self.assertEqual(stringify(None, str_none=True), str(None))



import bisect
import itertools
import math
from abc import ABCMeta, abstractmethod
import typing as tp

from satella.coding.concurrent.monitor import RMonitor
from satella.coding.recast_exceptions import silence_excs
from satella.coding.sequences import try_close
from satella.coding.typing import V, K, KVTuple


class DBStorage(metaclass=ABCMeta):
    """
    An abstract implementation of the storage class provided to
    :class:`~satella.coding.structures.SyncableDroppable`

    This serves as satella's hook into your database infrastructure.
    """
    __slots__ = ()

    @abstractmethod
    def put(self, key: K, value: V) -> None:
        """
        Put given value to storage at given key.

        This may block for a while.

        :param key: key to store
        :param value: value to store
        """

    @abstractmethod
    def iterate(self, starting_key: tp.Optional[K]) -> tp.Iterator[KVTuple]:
        """
        Return an iterator iterating from provided starting key to the end
        of the values, as read from the database.

        This may block for a while.

        This iterator will be closed upon no longer being necessary.

        :param starting_key: starting key, included, or None for iterate from the start
        :return: an iterator from provided key (included) to the end of the range
        """

    @abstractmethod
    def on_change_start_entry(self, start_entry: tp.Optional[K]) -> None:
        """
        Called by SyncableDroppable when starting entry (earliest entry encountered both in the DB
        and is memory) is changed.

        :param start_entry: new value of start entry or None if there are no entries at all
        """

    @abstractmethod
    def on_change_stop_entry(self, stop_entry: tp.Optional[K]) -> None:
        """
        Called by SyncableDroppable when stopping entry (earliest entry encountered both in the DB
        and is memory) is changed.

        :param stop_entry: new value of stop entry or None if there are no entries at all
        """

    @abstractmethod
    def on_change_synced_up_to(self, synced_up_to: tp.Optional[K]) -> None:
        """
        Called by SyncableDroppable when synced up to (earliest timestamp synced) is changed.

        :param synced_up_to: new value of synced up to
        """

    @abstractmethod
    def delete(self, key: K) -> None:
        """
        Called by SyncableDroppable when there's a need to remove target key

        :param key: key to remove
        """


class SyncableDroppable(RMonitor, tp.Generic[K, V]):
    """
    A thread-safe class representing some single time series, which needs to be synchronized with some server,
    and may be too large to keep in memory. Moreover, after the sync we need to retain a part of the
    time series in memory for future requests. Only series past some timestamp may be deleted.

    For brevity, this will refer to keys as timestamps. The keys must be __eq__able, comparable
    and subtractable.

    A rule is that an item can never be both in memory and in the DB.

    :param db_storage: a DBStorage implementation of your own provision, that serves as class'
        interface with the database
    :param start_entry: earliest timestamp stored or None if no data is stored
    :param stop_entry: latest timestamp stored or None if no data is stored
    :param synced_up_to: timestamp of the last entry synchronized or None if no data is stored
    :param span_to_keep_in_memory: key span to keep in memory. Entries earlier than
        difference of the latest key and this will be dropped from memory, onto the DB.
        Can't be false.
    :param span_to_keep_in_db: key span to keep on disk. Entries earlier than
        difference of the latest key and this will be dropped from the DB.
        Can't be false.

    .. note:: Note that proper handling of maximum spans requires periodical calls to
              :meth:`~satella.coding.structures.SyncableDroppable.cleanup`
    """
    __slots__ = ('db_storage', '_start_entry', '_stop_entry', '_synced_up_to', 'data_in_memory',
                 'span_to_keep_in_memory', 'span_to_keep_in_db')

    def __init__(self, db_storage: DBStorage, start_entry: tp.Optional[K],
                 stop_entry: tp.Optional[K], synced_up_to: tp.Optional[K],
                 span_to_keep_in_memory: int, span_to_keep_in_db: int):
        super().__init__()
        assert span_to_keep_in_memory and span_to_keep_in_db, 'One of spans was false!'
        assert span_to_keep_in_db > span_to_keep_in_memory, 'Invalid span'
        self.db_storage = db_storage  # type: DBStorage
        self._start_entry = start_entry  # type: K
        self._stop_entry = stop_entry  # type: K
        self._synced_up_to = synced_up_to  # type: K
        self.data_in_memory = []  # type: tp.List[KVTuple]
        self.span_to_keep_in_db = span_to_keep_in_db  # type: K
        self.span_to_keep_in_memory = span_to_keep_in_memory  # type: K

    @property
    def start_entry(self) -> tp.Optional[K]:
        return self._start_entry

    @start_entry.setter
    def start_entry(self, v: tp.Optional[K]) -> None:
        self._start_entry = v
        self.db_storage.on_change_start_entry(v)

    @property
    def synced_up_to(self) -> tp.Optional[K]:
        return self._synced_up_to

    @synced_up_to.setter
    def synced_up_to(self, v: tp.Optional[K]) -> None:
        self._synced_up_to = v
        self.db_storage.on_change_synced_up_to(v)

    @property
    def stop_entry(self) -> tp.Optional[K]:
        return self._stop_entry

    @stop_entry.setter
    def stop_entry(self, v: tp.Optional[K]) -> None:
        self._stop_entry = v
        self.db_storage.on_change_stop_entry(v)

    @RMonitor.synchronized
    def sync_to_db(self) -> None:
        """
        Make sure that everything's that in memory in also stored in the DB.
        """
        for key, value in self.data_in_memory:
            self.db_storage.put(key, value)
        self.data_in_memory = []

    def cleanup(self) -> None:
        """
        Make sure that everything's that in memory and the DB conforms to span_to_keep_in_db
        and span_to_keep_in_memory.

        This may block for a while.
        """
        self.cleanup_keep_in_db()
        self.cleanup_keep_in_memory()

    def _cleanup_the_db(self) -> bool:
        """
        Remove entries from the DB that are older than span_to_keep_in_db

        :return: if all entries in the DB have been trashed
        """
        if self.start_entry is None:
            return False
        cutoff_span = self.stop_entry - self.span_to_keep_in_db
        iterator = self.db_storage.iterate(self.start_entry)
        try:
            for key, value in iterator:
                if key < cutoff_span:
                    self.db_storage.delete(key)
                else:
                    self.start_entry = key
                    break
            else:
                # This means that we have wiped entire DB
                if self.data_in_memory:
                    self.start_entry = self.data_in_memory[0][0]
                    self.db_storage.on_change_start_entry(self.start_entry)
                else:
                    # We no longer have ANY data
                    self.start_entry = self.stop_entry = None
                return True

        finally:
            try_close(iterator)
        return False

    def get_archive(self, start: K, stop: K) -> tp.Iterator[KVTuple]:
        """
        Get some historic data that is kept both in the DB and in the memory

        :param start: starting key (included)
        :param stop: stopping key (included)
        :return: a iterator of KVTuple
        """
        if not self.data_in_memory:
            return []
        if self.first_key_in_memory <= start:
            # We'll serve it from memory
            for key, value in self.data_in_memory:
                if key < start:
                    continue
                if key > stop:
                    return
                yield key, value
        else:
            it = self.db_storage.iterate(start)
            try:
                for key, value in it:
                    if key < start:
                        continue
                    if key > stop:
                        return
                    yield key, value
                # We must iterate from the memory
                if self.data_in_memory:
                    yield from self.get_archive(self.first_key_in_memory, stop)
            finally:
                try_close(it)

    def get_latest_value(self) -> KVTuple:
        """
        Get the piece of data that was added here last

        :return: a tuple of (key, value)
        :raise ValueError: no data in series
        """
        if self.stop_entry is None:
            raise ValueError('No data in series')
        if self.data_in_memory:
            return self.data_in_memory[-1]
        else:
            iterator = self.db_storage.iterate(self.stop_entry)
            try:
                return next(iterator)
            finally:
                iterator.close()

    @RMonitor.synchronized
    def cleanup_keep_in_memory(self) -> None:
        """
        Eject values from memory that should reside in the DB onto the DB
        """
        first_key = self.first_key_in_memory
        if first_key is None:
            return
        cutoff_point = self.stop_entry - self.span_to_keep_in_memory
        for index, row in enumerate(self.data_in_memory):
            ts, value = row
            if ts > cutoff_point:
                for ts, value in self.data_in_memory[:index]:
                    self.db_storage.put(ts, value)
                del self.data_in_memory[:index]
                break
        else:
            self.sync_to_db()
            self.data_in_memory = []
            return

    @RMonitor.synchronized
    def cleanup_keep_in_db(self) -> None:
        """
        Clear up the database to conform to our span_to_keep_in_db
        """
        if self.start_entry is None or not self.data_in_memory:
            return
        cutoff_span = self.stop_entry - self.span_to_keep_in_db
        if self.start_entry == self.first_key_in_memory:
            # The entire series is loaded in the memory
            self.data_in_memory = [tpl for tpl in self.data_in_memory if tpl[0] >= cutoff_span]
            if self.data_in_memory:
                self.start_entry = self.first_key_in_memory
            else:
                self.start_entry = self.stop_entry = None
        else:
            if not self._cleanup_the_db():
                return
            self.cleanup_keep_in_db()

    def on_new_data(self, key: K, value: V) -> None:
        """
        Called by the user when there's new data gathered.

        Key must be greater than start entry

        :param key: key of the new data
        :param value: value of the new data
        :raise ValueError: key was not larger than current stop entry
        """
        if self.stop_entry is not None:
            if key <= self.stop_entry:
                raise ValueError('Key not greater than current stop entry!')

        self.data_in_memory.append((key, value))
        self.stop_entry = key
        if self.start_entry is None:
            self.start_entry = key

    def on_sync_request(self, maximum_entries: tp.Optional[int] = math.inf) -> tp.Iterator[KVTuple]:
        """
        Return an iterator that will provide the source of the data for synchronization.

        This will preferentially start from the first value, so as to keep values synchronized
        in-order.

        :param maximum_entries:
        :return: an iterator of (KVTuple) that should be synchronized against the server
        :raise ValueError: nothing to synchronize!
        """
        if not self.start_entry:
            raise ValueError('Nothing to synchronize!')
        if self.synced_up_to == self.data_in_memory[-1][0]:
            raise ValueError('Nothing to synchronize!')
        if self.synced_up_to is None:
            # Sync everything
            iterator = self.db_storage.iterate(None)
            try:
                data = list(iterator)
                if len(data) < maximum_entries:
                    entries_left = maximum_entries - len(data)
                    if entries_left == math.inf:
                        data = itertools.chain(data, self.data_in_memory)
                    else:
                        data = itertools.chain(data, self.data_in_memory[:entries_left])
                v = data
            finally:
                try_close(iterator)
        else:
            if self.first_key_in_memory <= self.synced_up_to:
                # Means we have to sync from memory
                if self.synced_up_to is None:
                    v = self.data_in_memory
                else:
                    index = bisect.bisect_right([y[0] for y in self.data_in_memory],
                                                self.synced_up_to)
                    if maximum_entries == math.inf:
                        v = self.data_in_memory[index:]
                    else:
                        v = self.data_in_memory[index:index + maximum_entries]
            else:
                # We have to start off the disk
                data = []
                iterator = self.db_storage.iterate(self.start_entry)
                try:
                    while len(data) < maximum_entries:
                        try:
                            data.append(next(iterator))
                        except StopIteration:
                            for index, tpl in enumerate(self.data_in_memory):
                                if len(data) >= maximum_entries:
                                    break
                                if self.synced_up_to is not None:
                                    if tpl[0] > self.synced_up_to:
                                        break
                            v = itertools.chain(data, self.data_in_memory[:index])
                            break
                    else:
                        v = data
                finally:
                    try_close(iterator)
        return v

    def on_synced_up_to(self, key: K) -> None:
        """
        Called when data was successfully synced up to key included

        :param key: maximum key synchronized
        """
        self.synced_up_to = key

    @property
    def first_key_in_memory(self) -> tp.Optional[K]:
        """
        :return: key of the first element stored in memory
        """
        if not self.data_in_memory:
            return None
        else:
            return self.data_in_memory[0][0]



import enum
import os

from satella.configuration.sources import load_source_from_dict, load_source_from_list
from .test_sources.utils import SourceTestCase

INNER_DATA = [
    {
        'type': 'BuildObjectFrom',
        'key': 'test',
        'child': {
            'type': 'StaticSource',
            'args': [
                {
                    "a": 5
                }
            ]
        }
    },
    {
        'type': 'JSONSource',
        'arg': '{"d": 10}'
    },
    {
        'type': 'JSONSource',
        'args': ['{"a": 5}']
    },
    {
        'type': 'JSONSource',
        'args': ['ivnadsgfgsfg'],
        'optional': True
    },
    {
        'type': 'JSONSource',
        'args': ['{"b": 5}']
    },
    {
        'type': 'JSONSource',
        'encoding': "base64",
        'args': [
            {
                'type': 'binary',
                'value': 'eyJjIjogMjF9\n',
            }
        ]
    }
]

INNER_DATA_2 = {
    'type': 'EnvironmentSource',
    'args': [
        'TEST_ENV', 'TEST_ENV'
    ],
    'cast_to': {
        'type': 'import',
        'module': 'tests.test_configuration.test_load_source_from_dict',
        'attribute': 'MyEnum',
        'cast_before': {
            'type': 'lambda',
            'operation': 'int(x)'
        }
    }
}


class MyEnum(enum.IntEnum):
    A = 0

DICT_DATA = {
    'type': 'MergingSource',
    'args': INNER_DATA
}


class TestLoadSourceFromDict(SourceTestCase):
    def test_lsf(self):
        output = {'a': 5, 'b': 5, 'c': 21, 'test': {'a': 5}, 'd': 10}
        self.assertSourceHas(load_source_from_dict(DICT_DATA), output)
        self.assertSourceHas(load_source_from_list(INNER_DATA), output)

    def test_env(self):
        os.environ['TEST_ENV'] = '0'
        source = load_source_from_dict(INNER_DATA_2).provide()
        self.assertEqual(source['TEST_ENV'], MyEnum(0))



import sys


# shamelessly stolen from
# https://goshippo.com/blog/measure-real-size-any-python-object/
def get_size(obj, seen=None) -> int:
    """
    Recursively finds the total size of an object (object + it's components).

    :param obj: object to measure
    :return: size in bytes of the object and all of it's subcomponents
    :raises RuntimeError: when ran on PyPy
    """
    try:
        size = sys.getsizeof(obj)
    except TypeError:
        raise RuntimeError('Running on PyPy?')
    if seen is None:
        seen = set()
    obj_id = id(obj)
    if obj_id in seen:
        return 0
    # Important mark as seen *before* entering recursion to gracefully handle
    # self-referential objects
    seen.add(obj_id)
    if isinstance(obj, dict):
        size += sum(get_size(v, seen) for v in obj.values())
        size += sum(get_size(k, seen) for k in obj.keys())
    elif hasattr(obj, '__dict__'):
        size += get_size(obj.__dict__, seen)
    elif hasattr(obj, '__iter__') and not isinstance(obj, (str, bytes, bytearray)):
        size += sum(get_size(i, seen) for i in obj)
    return size



from satella.configuration.sources import MergingSource, JSONSource, \
    AlternativeSource, StaticSource
from satella.exceptions import ConfigurationError
from .utils import SourceTestCase


class TestMergingSource(SourceTestCase):
    def test_ok(self):
        s = MergingSource(
            JSONSource('{"a": [5]}'),
            JSONSource('{"a": [6]}'),
        )
        self.assertSourceHas(s, {"a": [5, 6]})

    def test_fails_on_none(self):
        s = MergingSource(
            JSONSource('{"a: [5]}'),
            on_fail=MergingSource.SILENT,
            fail_if_no_sources_are_correct=True
        )
        self.assertRaises(ConfigurationError, s.provide)

        s = MergingSource(
            JSONSource('{"a: [5]}'),
            on_fail=MergingSource.SILENT,
            fail_if_no_sources_are_correct=False
        )
        self.assertEqual(s.provide(), {})

    def test_empty(self):
        self.assertSourceEmpty(MergingSource())

    def test_static(self):
        st = StaticSource({'a': 5})
        self.assertEqual(st.provide(), {'a': 5})


class TestAlternativeSource(SourceTestCase):
    def test_alternative(self):
        self.assertSourceFails(AlternativeSource(), ConfigurationError)



from .base import BaseSource

__all__ = ['BuildObjectFrom']


class BuildObjectFrom(BaseSource):
    """
    A source that outputs a single key with given name, and as it's contents the contents
    of it's child.
    """
    __slots__ = 'key', 'child'

    def __init__(self, key: str, child: BaseSource):
        self.key = key
        self.child = child

    def provide(self) -> dict:
        """
        Return your configuration, as a dict

        :raise ConfigurationError: on invalid configuration
        """
        return {self.key: self.child.provide()}

    def __repr__(self) -> str:
        return '<%s>' % (self.__class__.__qualname__,)



import enum
import os
import shutil
import tempfile
import unittest

from satella.configuration.schema import *
from satella.configuration.sources import DirectorySource
from satella.exceptions import ConfigurationValidationError


class Environment(enum.IntEnum):
    PRODUCTION = 0


class TestSchema(unittest.TestCase):

    def test_file(self):
        schema = {
            "key": "file"
        }
        with open('test', 'wb') as f_out:
            f_out.write(b'test')

        s = descriptor_from_dict(schema)
        fo = s({'key': "test"})['key']
        self.assertEqual(str(fo), 'test')
        self.assertEqual(fo.get_value(), b'test')
        self.assertEqual(fo.get_value('utf-8'), 'test')

    def test_file_contents_1(self):
        schema = {
            "key": {
                "type": "file_contents",
                "encoding": "utf-8"
            }
        }
        with open('test', 'wb') as f_out:
            f_out.write(b'test')

        s = descriptor_from_dict(schema)
        fo = s({'key': "test"})['key']
        self.assertEqual(fo, 'test')

    def test_file_contents_2(self):
        schema = {
            "key": {
                "type": "file_contents",
                "strip_afterwards": True
            }
        }
        with open('test', 'wb') as f_out:
            f_out.write(b'test\n')

        s = descriptor_from_dict(schema)
        fo = s({'key': "test"})['key']
        self.assertEqual(fo, b'test')

    def test_caster(self):
        ps = Caster(Environment)
        self.assertEqual(ps(0), Environment.PRODUCTION)

    def test_directory(self):
        schema = {
            "directory": "dir"
        }

        if os.path.exists('directory'):
            shutil.rmtree('directory')

        s = descriptor_from_dict(schema)
        self.assertRaises(ConfigurationValidationError, lambda: s({'directory': 'directory'}))
        os.mkdir('directory')
        self.assertEqual(s({'directory': 'directory'})['directory'], DirectoryObject('directory'))

    def test_descriptor_from_schema_caster(self):
        schema = {
            "key": {
                'type': 'caster',
                'cast_to': 'tests.test_configuration.test_schema.Environment',
                'expr': 'y(int(x))'
            }
        }

        s = descriptor_from_dict(schema)

        self.assertEqual(s({'key': "0"}), {'key': Environment.PRODUCTION})

    def test_union(self):
        ps = Union(List(), Dict(keys=[create_key(String(), 'a')]))

        self.assertEqual(ps([1, 2, 3]), [1, 2, 3])
        self.assertEqual(ps({'a': 'b'}), {'a': 'b'})
        self.assertRaises(ConfigurationValidationError, lambda: ps(3))

    def test_descriptor_from_schema(self):
        schema = {
            "key_s": "str",
            "key_i": {
                "type": "int",
                "description": "Value of key I"
            },
            "key_f": "float",
            "ip_addr": "ipv4",
            "nested": {
                "key_s": "str",
            },
            "default_five": {
                "type": "int",
                "default": 5
            },
            'logstash': {
                'type': 'list',
                'of': {
                    'host': 'str',
                    'port': {
                        'type': 'int',
                        'default': 9600,
                        'optional': True
                    }
                }
            },
            'boolean2': 'bool',
            'boolean': 'bool',
            'anything': 'any',
            "list": {
                "type": "list",
                "of": "str"
            },
            "unioned_list": {
                "type": "list",
                "of": {
                    "type": "union",
                    "of": [
                        {
                            "a": "str"
                        },
                        "int"
                    ]
                }}
        }

        s = descriptor_from_dict(schema)
        self.assertEqual(s({
            'key_s': 'string',
            'key_i': '5',
            'key_f': '5.5',
            'ip_addr': '10.2.3.43',
            'anything': 'any',
            'boolean': 'false',
            'unioned_list': [{"a": "a"}, 5],
            'nested': {
                'key_s': "string"
            },
            'boolean2': 0,
            'logstash': [
                {'host': 'ipv4'}
            ],
            'list': [1, 2, 3]
        }), {
            'key_s': 'string',
            'key_i': 5,
            'key_f': 5.5,
            'boolean': False,
            'ip_addr': '10.2.3.43',
            'anything': 'any',
            'boolean2': False,
            'unioned_list': [{"a": "a"}, 5],
            'nested': {
                'key_s': "string"
            },
            'logstash': [{'host': 'ipv4', 'port': 9600}],
            'list': ['1', '2', '3'],
            'default_five': 5
        })

    def test_optional_list(self):
        d1 = {
            'files': {
                'type': 'list',
                'of': {
                    'name': 'str',
                    'content': 'str'
                },
                'default': []
            }
        }
        d = descriptor_from_dict(d1)({})
        self.assertIn('files', d)

    def test_schema(self):
        D1 = {
            'key_s': 'value',
            'key_i': '5',
            'key_f': '5.2',
            'unknown_key': None,
            'ip_addr': '127.0.0.1'
        }

        s = Dict([
            create_key(String(), 'key_s'),
            create_key(Integer(), 'key_i'),
            create_key(Float(), 'key_f'),
            create_key(String(), 'key_not_present', optional=True,
                       default='hello world'),
            create_key(IPv4(), 'ip_addr')
        ], unknown_key_mapper=lambda key, value: str(value))

        D2 = D1.copy()
        D2.update(key_not_present='hello world', key_i=5, key_f=5.2,
                  unknown_key='None')
        self.assertEqual(s(D1), D2)

    def test_schema_x(self):
        dir = tempfile.mkdtemp()

        with open(os.path.join(dir, 'smok5_config.json'), 'w') as f_out:
            f_out.write("""{
  "logging": {
    "logstash": {
      "host": "192.168.10.11",
      "port": 5959
    }
  }
}""")
        ds = DirectorySource(dir)
        source = ds.provide()

        schema = Dict([
            create_key(Dict([
                create_key(Dict([
                    create_key(String(), 'host'),
                    create_key(Integer(), 'port')
                ]), 'logstash')
            ]), 'logging')
        ])
        source = schema(source)
